//
// JTypist v3.1 - Java typing tutor program for UNIX systems
// Copyright (C) 2000  Simon Baldwin (simonb@sco.com)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//
// Tooltip class - displays tool tips for an AWT components
//
import java.awt.Window;
import java.awt.Graphics;
import java.awt.Component;
import java.awt.Frame;
import java.awt.Dimension;
import java.awt.Panel;
import java.awt.BorderLayout;
import java.awt.Point;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Button;
import java.awt.event.MouseListener;
import java.awt.event.MouseEvent;
import java.awt.event.KeyListener;
import java.awt.event.KeyEvent;
import java.awt.IllegalComponentStateException;
import java.util.ResourceBundle;

/**
 * Displays an AWT component tooltip.
 */
public class Tooltip extends Window
			implements Runnable, MouseListener, KeyListener {

	// message and colour catalog information
	private static final ResourceBundle	messages	=
				ResourceBundle.getBundle( "MessageResource" );
	private static final ResourceBundle	colours		=
				ResourceBundle.getBundle( "ColorResource" );

	// flag to indicate if we should not attempt any tooltips at all
	private static boolean			enabled	= true;

	// delay before displaying a tooltip
	private static final long	POPUP_DELAY	=
					(long)(Integer.parseInt(
						messages.getString(
							"Tooltip.delay" )));

	// padding round text in the tooltip window, and X/Y offset
	private static final int	PADDING		=
					Integer.parseInt(
						messages.getString(
							"Tooltip.padding" ));
	private static final int	XOFFSET		=
					Integer.parseInt(
						messages.getString(
							"Tooltip.xoffset" ));
	private static final int	YOFFSET		=
					Integer.parseInt(
						messages.getString(
							"Tooltip.yoffset" ));

	// thread for implementing delays; here we cheap out, and make all
	// tooltip instances share a single thread.  We can get away with
	// this because in practice, the mouse can only be over one component
	// at any time.  Plus, if we didn't do this, there would be an awful
	// lot of threads going.
	private static Thread		thread;//	= null;

	// tooltip currently awaiting a callback; null if none pending
	private static Tooltip		cb_tooltip;//	= null;

	// tooltip message string
	private String			message;//	= null;


	/**
	 * Create a new Tooltip.
	 */
	public Tooltip( Frame parent, Component component, String message ) {
		super( parent );

		// note the message
		this.message = message;

		// set the colours for tooltips
		setForeground( Colour.find_colour(
			colours.getString( "Tooltip.foreground" )));
		setBackground( Colour.find_colour(
			colours.getString( "Tooltip.background" )));
		setFont( Font.decode(
			colours.getString( "Tooltip.font" )));

		// size the window for the given font
		FontMetrics fm = getFontMetrics( getFont() );
		setSize( fm.stringWidth( message ) + PADDING,
						fm.getAscent() + PADDING );

		// set ourselves up to receive mouse and key events
		component.addMouseListener( this );
		component.addKeyListener( this );

		// if no timer thread is yet running, then consider ourself
		// as the master for the class by starting one
		if ( thread == null ) {
			thread		= new Thread( this, "Tooltip_timer" );
			thread.start();
		}
	}

	/**
	 * Paint the graphic.
	 */
	public void paint( Graphics g ) {
	synchronized( thread ) {
		Dimension d = getSize();

		// draw a border
		g.setColor( getForeground() );
		g.drawRect( 0, 0, d.width - 1, d.height - 1 );

		// display the message text
		FontMetrics fm	= g.getFontMetrics();
		g.drawString( message, PADDING / 2,
				( d.height + fm.getAscent() ) / 2 );
	}}

	/**
	 * Thread to implement delays on tooltip appearance.
	 */
	public void run() {
	synchronized( thread ) {

		// loop forever, at the lowest priority
		Thread.currentThread().setPriority( Thread.MIN_PRIORITY );
		while ( true ) {

			// wait until the timeout expires, or if no request is
			// pending, just wait
			try {
				// see if there is a request pending
				if ( cb_tooltip != null ) {

					// wait until it's time to callback
					thread.wait( POPUP_DELAY );

					// sleep completed; call the callback
					if ( cb_tooltip != null ) {
						cb_tooltip.callback();

						// delete the callback request
						cb_tooltip = null;
					}
				}
				else {
					// no request, so simply wait
					thread.wait();
				}
			}
			catch ( InterruptedException e ) {
				// probably a change of callback object;
				// ignore, and simply loop
			}
		}
	}}

	/**
	 * Enable or disable ALL tooltips.
	 */
	public static void set_enabled( boolean new_enabled ) {
		enabled = new_enabled;
	}
	public static boolean is_enabled() {
		return ( enabled );
	}

	/*
	 * Request, and cancel, a callback.
	 */
	private void request_callback() {
	synchronized( thread ) {
		// always write, or overwrite, with our object
		cb_tooltip	= this; thread.interrupt();
	}}
	private void cancel_callback() {
	synchronized( thread ) {
		// only clear the request if it is ours to clear
		if ( cb_tooltip == this ) {
			cb_tooltip	= null; thread.interrupt();
		}
	}}

	/**
	 * Event handlers.
	 */
	public void mouseClicked( MouseEvent e ) {}
	public void mousePressed( MouseEvent e ) {
		cancel_callback(); setVisible( false ); repaint();
	}
	public void mouseReleased( MouseEvent e ) {}
	public void mouseEntered( MouseEvent e ) {

		// ignore the call if tooltips are not enabled
		if ( enabled ) {

			// use try-catch to avoid illegal component state that
			// can occur when this thread races with the graphics
			// updater; we can't get the screen location if the
			// component happens not to be showing
			Point p = null;
			try {
				p = e.getComponent().getLocationOnScreen();
			}
			catch ( IllegalComponentStateException exception ) {
				return;
			}
			setLocation( p.x + e.getX() + XOFFSET,
						p.y + e.getY() + YOFFSET );

			// request a callback after a short delay
			request_callback();
		}
	}
	public void callback() {
		setVisible( true ); repaint();
	}
	public void mouseExited( MouseEvent e ) {
		cancel_callback(); setVisible( false ); repaint();
	}
	public void keyTyped( KeyEvent evt ) {}
	public void keyPressed( KeyEvent evt ) {
		cancel_callback(); setVisible( false ); repaint();
	}
	public void keyReleased( KeyEvent evt ) {}

	/**
	 * Class test code.
	 */
	public static void main( String args[] ) {
		System.out.println( "*** CLASS TEST CODE ***" );

		Frame frame = new Frame();
		Button b1 = new Button( "Testbutton1" );
		Button b2 = new Button( "Testbutton2" );
		Tooltip tip1 = new Tooltip( frame, b1, "Tip string 1" );
		Tooltip tip2 = new Tooltip( frame, b2, "Tip string 2" );
		Panel p = new Panel();
		p.add( b1 ); p.add( b2 );
		frame.add( p );
		frame.setSize( 100, 100 );
		frame.setVisible( true );
	}
}
